@@ -28,6 +28,7 @@
 #include <hurd/iohelp.h>
 #include <idvec.h>
 #include <features.h>
+#include <refcount.h>
 
 #ifdef DISKFS_DEFINE_EXTERN_INLINE
 #define DISKFS_EXTERN_INLINE
@@ -55,9 +56,9 @@ struct protid
 /* One of these is created for each node opened by dir_lookup. */
 struct peropen
 {
-  int filepointer;
+  off_t filepointer;
   int lock_status;
-  int refcnt;
+  refcount_t refcnt;
   int openstat;
 
   struct node *np;
@@ -79,24 +80,31 @@ struct peropen
    filesystem.  */
 struct node
 {
-  struct node *next, **prevp;
+  /* Links on hash list. */
+  struct node *hnext, **hprevp;
 
   struct disknode *dn;
 
   io_statbuf_t dn_stat;
 
+  /* Flags.  */
+  unsigned int
+
   /* Stat has been modified if one of the following four fields
      is nonzero.  Also, if one of the dn_set_?time fields is nonzero,
      the appropriate dn_stat.st_?tim field needs to be updated. */
-  int dn_set_ctime;
-  int dn_set_atime;
-  int dn_set_mtime;
-  int dn_stat_dirty;
+    dn_set_ctime:1,
+    dn_set_atime:1,
+    dn_set_mtime:1,
+    dn_stat_dirty:1,
+
+  /* Indicate whether the author is tracking the uid because the
+     on-disk file format does not encode a separate author.  */
+    author_tracks_uid:1;
 
   pthread_mutex_t lock;
 
-  int references;		/* hard references */
-  int light_references;		/* light references */
+  refcounts_t refcounts;
 
   mach_port_t sockaddr;		/* address for S_IFSOCK shortcut */
 
@@ -117,8 +125,16 @@ struct node
   loff_t allocsize;
 
   ino64_t cache_id;
+};
 
-  int author_tracks_uid;
+struct diskfs_control
+{
+  struct port_info pi;
+};
+
+struct bootinfo
+{
+  struct port_info pi;
 };
 
 /* Possibly lookup types for diskfs_lookup call */
@@ -187,8 +203,6 @@ extern volatile struct mapped_time_value *diskfs_mtime;
    be done by format independent code. */
 extern int diskfs_synchronous;
 
-extern pthread_spinlock_t diskfs_node_refcnt_lock;
-
 extern int pager_port_type;
 
 /* Whether the filesystem is currently writable or not. */
@@ -283,7 +297,7 @@ int diskfs_shortcut_ifsock;
 
 /* The user may define this variable, otherwise it has a default value of 30.
    diskfs_set_sync_interval is called with this value when the first diskfs
-   thread is started up (in diskfs_spawn_first_threa).   */
+   thread is started up (in diskfs_spawn_first_thread).   */
 extern int diskfs_default_sync_interval;
 
 /* The user must define this variable, which should be a string that somehow
@@ -437,14 +451,16 @@ error_t diskfs_alloc_node (struct node *dp, mode_t mode, struct node **np);
 void diskfs_free_node (struct node *np, mode_t mode);
 
 /* Node NP has no more references; free local state, including *NP
-   if it isn't to be retained.  diskfs_node_refcnt_lock is held. */
+   if it isn't to be retained.  */
 void diskfs_node_norefs (struct node *np);
 
-/* The user must define this function.  Node NP has some light
+/* The user must define this function unless she wants to use the node
+   cache.  See the section `Node cache' below.  Node NP has some light
    references, but has just lost its last hard references.  Take steps
    so that if any light references can be freed, they are.  NP is locked
    as is the pager refcount lock.  This function will be called after
-   diskfs_lost_hardrefs.  */
+   diskfs_lost_hardrefs.  An additional light reference is acquired by
+   libdiskfs across calls to this function.  */
 void diskfs_try_dropping_softrefs (struct node *np);
 
 /* The user must define this funcction.  Node NP has some light
@@ -503,7 +519,8 @@ void diskfs_write_disknode (struct node *np, int wait);
    then return only after the physical media has been completely updated.  */
 void diskfs_file_update (struct node *np, int wait);
 
-/* The user must define this function.  For each active node, call
+/* The user must define this function unless she wants to use the node
+   cache.  See the section `Node cache' below.  For each active node, call
    FUN.  The node is to be locked around the call to FUN.  If FUN
    returns non-zero for any node, then immediately stop, and return
    that value. */
@@ -569,10 +586,41 @@ error_t (*diskfs_create_symlink_hook)(struct node *np, const char *target);
 error_t (*diskfs_read_symlink_hook)(struct node *np, char *target);
 
 /* The user may define this function.  The function must set source to
-   the source device of the filesystem. The function may return an
-   EOPNOTSUPP to indicate that the concept of a source device is not
-   applicable. The default function always returns EOPNOTSUPP. */
-error_t diskfs_get_source (char *source);
+   the source of CRED. The function may return an EOPNOTSUPP to
+   indicate that the concept of a source device is not applicable. The
+   default function always returns EOPNOTSUPP. */
+error_t diskfs_get_source (struct protid *cred,
+                           char *source, size_t source_len);
+
+/* Libdiskfs contains a node cache.
+
+   Using it relieves the user of implementing diskfs_cached_lookup,
+   diskfs_node_iterate, and diskfs_try_dropping_softrefs.
+
+   In order to use it, she must implement the following functions with
+   the prefix `diskfs_user_'.  */
+
+/* This can be used to provide additional context to
+   diskfs_user_make_node and diskfs_user_read_node in case of cache
+   misses.  */
+struct lookup_context;
+
+/* The user must define this function if she wants to use the node
+   cache.  Create and initialize a node.  */
+error_t diskfs_user_make_node (struct node **npp, struct lookup_context *ctx);
+
+/* The user must define this function if she wants to use the node
+   cache.  Read stat information out of the on-disk node.  */
+error_t diskfs_user_read_node (struct node *np, struct lookup_context *ctx);
+
+/* The user must define this function if she wants to use the node
+   cache.  The last hard reference to a node has gone away; arrange to
+   have all the weak references dropped that can be.  */
+void diskfs_user_try_dropping_softrefs (struct node *np);
+
+/* Lookup node INUM (which must have a reference already) and return it
+   without allocating any new references. */
+struct node *diskfs_cached_ifind (ino_t inum);
 
 /* The library exports the following functions for general use */
 
@@ -599,9 +647,8 @@ void diskfs_spawn_first_thread (ports_demuxer_type demuxer);
    diskfs_init_completed once it has a valid proc and auth port. */
 void diskfs_start_bootstrap ();
 
-/* Node NP now has no more references; clean all state.  The
-   _diskfs_node_refcnt_lock must be held, and will be released
-   upon return.  NP must be locked.  */
+/* Node NP now has no more references; clean all state.  NP must be
+   locked.  */
 void diskfs_drop_node (struct node *np);
 
 /* Set on disk fields from NP->dn_stat; update ctime, atime, and mtime
@@ -674,6 +721,45 @@ diskfs_notice_filechange (struct node *np, enum file_changed_type type,
    The new node will have one hard reference and no light references.  */
 struct node *diskfs_make_node (struct disknode *dn);
 
+/* Create a new node structure.  Also allocate SIZE bytes for the
+   disknode.  The address of the disknode can be obtained using
+   diskfs_node_disknode.  The new node will have one hard reference
+   and no light references.  */
+struct node *diskfs_make_node_alloc (size_t size);
+
+/* To avoid breaking the ABI whenever sizeof (struct node) changes, we
+   explicitly provide the size.  The following two functions will use
+   this value for offset calculations.  */
+extern const size_t _diskfs_sizeof_struct_node;
+
+/* Return the address of the disknode for NODE.  NODE must have been
+   allocated using diskfs_make_node_alloc.  */
+struct disknode *diskfs_node_disknode (struct node *node);
+
+/* Return the address of the node for DISKNODE.  DISKNODE must have
+   been allocated using diskfs_make_node_alloc.  */
+struct node *diskfs_disknode_node (struct disknode *disknode);
+
+#if defined(__USE_EXTERN_INLINES) || defined(DISKFS_DEFINE_EXTERN_INLINE)
+
+/* Return the address of the disknode for NODE.  NODE must have been
+   allocated using diskfs_make_node_alloc.  */
+DISKFS_EXTERN_INLINE struct disknode *
+diskfs_node_disknode (struct node *node)
+{
+  return (struct disknode *) ((char *) node + _diskfs_sizeof_struct_node);
+}
+
+/* Return the address of the node for DISKNODE.  DISKNODE must have
+   been allocated using diskfs_make_node_alloc.  */
+DISKFS_EXTERN_INLINE struct node *
+diskfs_disknode_node (struct disknode *disknode)
+{
+  return (struct node *) ((char *) disknode - _diskfs_sizeof_struct_node);
+}
+
+#endif /* Use extern inlines.  */
+
 
 /* The library also exports the following functions; they are not generally
    useful unless you are redefining other functions the library provides. */
@@ -765,9 +851,18 @@ error_t diskfs_dirrewrite (struct node *dp, struct node *oldnp,
 error_t diskfs_dirremove (struct node *dp, struct node *np,
 			  const char *name, struct dirstat *ds);
 
-/* Return the node corresponding to CACHE_ID in *NPP. */
+/* The user must define this function unless she wants to use the node
+   cache.  See the section `Node cache' above.  Return the node
+   corresponding to CACHE_ID in *NPP. */
 error_t diskfs_cached_lookup (ino64_t cache_id, struct node **npp);
 
+/* Return the node corresponding to CACHE_ID in *NPP.  In case of a
+   cache miss, use CTX to create it and load it from the disk.  See
+   the section `Node cache' above.  */
+error_t diskfs_cached_lookup_context (ino_t inum, struct node **npp,
+				      struct lookup_context *ctx);
+
+
 /* Create a new node. Give it MODE; if that includes IFDIR, also
    initialize `.' and `..' in the new directory.  Return the node in NPP.
    CRED identifies the user responsible for the call.  If NAME is nonzero,
@@ -781,12 +876,12 @@ diskfs_create_node (struct node *dir, const char *name, mode_t mode,
 		    struct dirstat *ds);
 
 /* Create and return a protid for an existing peropen PO in CRED,
-   referring to user USER.  The node PO->np must be locked. */
+   referring to user USER.  On success, consume a reference to PO.  */
 error_t diskfs_create_protid (struct peropen *po, struct iouser *user,
 			      struct protid **cred);
 
 /* Build and return in CRED a protid which has no user identification, for
-   peropen PO.  The node PO->np must be locked.  */
+   peropen PO.  On success, consume a reference to PO.  */
 error_t diskfs_start_protid (struct peropen *po, struct protid **cred);
 
 /* Finish building protid CRED started with diskfs_start_protid;
@@ -794,8 +889,18 @@ error_t diskfs_start_protid (struct peropen *po, struct protid **cred);
 void diskfs_finish_protid (struct protid *cred, struct iouser *user);
 
 extern struct protid * diskfs_begin_using_protid_port (file_t port);
+extern struct protid *
+diskfs_begin_using_protid_payload (unsigned long payload);
+extern struct diskfs_control * diskfs_begin_using_control_port (fsys_t port);
+extern struct diskfs_control *
+diskfs_begin_using_control_port_payload (unsigned long payload);
+extern struct bootinfo *diskfs_begin_using_bootinfo_port (exec_startup_t port);
+struct bootinfo *
+diskfs_begin_using_bootinfo_payload (unsigned long payload);
 
 extern void diskfs_end_using_protid_port (struct protid *cred);
+extern void diskfs_end_using_control_port (struct diskfs_control *cred);
+extern void diskfs_end_using_bootinfo (struct bootinfo *upt);
 
 #if defined(__USE_EXTERN_INLINES) || defined(DISKFS_DEFINE_EXTERN_INLINE)
 
@@ -808,6 +913,44 @@ diskfs_begin_using_protid_port (file_t port)
   return ports_lookup_port (diskfs_port_bucket, port, diskfs_protid_class);
 }
 
+DISKFS_EXTERN_INLINE struct protid *
+diskfs_begin_using_protid_payload (unsigned long payload)
+{
+  return ports_lookup_payload (diskfs_port_bucket,
+			       payload,
+			       diskfs_protid_class);
+}
+
+/* And for the fsys interface. */
+DISKFS_EXTERN_INLINE struct diskfs_control *
+diskfs_begin_using_control_port (fsys_t port)
+{
+  return ports_lookup_port (diskfs_port_bucket, port, NULL);
+}
+
+DISKFS_EXTERN_INLINE struct diskfs_control *
+diskfs_begin_using_control_port_payload (unsigned long payload)
+{
+  return ports_lookup_payload (diskfs_port_bucket,
+			       payload,
+			       NULL);
+}
+
+/* And for the exec_startup interface. */
+DISKFS_EXTERN_INLINE struct bootinfo *
+diskfs_begin_using_bootinfo_port (exec_startup_t port)
+{
+  return ports_lookup_port (diskfs_port_bucket, port, diskfs_execboot_class);
+}
+
+DISKFS_EXTERN_INLINE struct bootinfo *
+diskfs_begin_using_bootinfo_payload (unsigned long payload)
+{
+  return ports_lookup_payload (diskfs_port_bucket,
+			       payload,
+			       diskfs_execboot_class);
+}
+
 /* Called by MiG after server routines have been run; this
    balances begin_using_protid_port, and is arranged for the io
    and fs interfaces by fsmutations.h. */
@@ -818,6 +961,22 @@ diskfs_end_using_protid_port (struct protid *cred)
     ports_port_deref (cred);
 }
 
+/* And for the fsys interface. */
+DISKFS_EXTERN_INLINE void
+diskfs_end_using_control_port (struct diskfs_control *cred)
+{
+  if (cred)
+    ports_port_deref (cred);
+}
+
+/* And for the exec_startup interface. */
+DISKFS_EXTERN_INLINE void
+diskfs_end_using_bootinfo (struct bootinfo *b)
+{
+  if (b)
+    ports_port_deref (b);
+}
+
 #endif /* Use extern inlines.  */
 
 /* Called when a protid CRED has no more references.  (Because references\
