@@ -142,6 +142,7 @@ pager_clear_user_data (struct user_pager_info *upi)
 }
 
 static struct port_bucket *pager_port_bucket = 0;
+static struct pager_requests *pager_requests;
 
 /* Initialize paging for this device.  */
 static void
@@ -160,7 +161,7 @@ init_dev_paging ()
 	  pager_port_bucket = ports_create_bucket ();
 
 	  /* Start libpagers worker threads.  */
-	  err = pager_start_workers (pager_port_bucket);
+	  err = pager_start_workers (pager_port_bucket, &pager_requests);
 	  if (err)
 	    {
 	      errno = err;
