@@ -36,6 +36,7 @@
 
 #include "libtrivfs/trivfs_fs_S.h"
 #include "libtrivfs/trivfs_io_S.h"
+#include "device_reply_S.h"
 
 /* The global lock */
 pthread_mutex_t global_lock;
@@ -273,10 +274,16 @@ static const struct argp argp = { options, parse_opt, args_doc, doc };
 int
 demuxer (mach_msg_header_t *inp, mach_msg_header_t *outp)
 {
-  extern int device_reply_server (mach_msg_header_t *, mach_msg_header_t *);
-
-  return (trivfs_demuxer (inp, outp)
-	  || device_reply_server (inp, outp));
+  mig_routine_t routine;
+  if ((routine = NULL, trivfs_demuxer (inp, outp)) ||
+      (routine = device_reply_server_routine (inp)))
+    {
+      if (routine)
+        (*routine) (inp, outp);
+      return TRUE;
+    }
+  else
+    return FALSE;
 }
 
 int
